@@ -1,6 +1,8 @@
 import copy
 from collections import defaultdict
 from functools import reduce, partial
+import itertools
+import re
 from typing import (
   Any, Dict, Optional, 
   Union, Tuple, cast,
@@ -18,11 +20,15 @@ from cloudvolume import CloudVolume
 from cloudvolume.lib import Vec, Bbox, max2, min2, xyzrange, find_closest_divisor, yellow, jsonify
 from cloudvolume.datasource.precomputed.sharding import ShardingSpecification
 from cloudfiles import CloudFiles
+import cloudfiles.paths
 
 from igneous.tasks import (
-  MeshTask, MeshManifestTask, GrapheneMeshTask,
+  MeshTask, MeshManifestPrefixTask, 
+  MeshManifestFilesystemTask, GrapheneMeshTask,
   MeshSpatialIndex, MultiResShardedMeshMergeTask,
-  MultiResUnshardedMeshMergeTask
+  MultiResUnshardedMeshMergeTask, 
+  MultiResShardedFromUnshardedMeshMergeTask,
+  TransferMeshFilesTask
 )
 from .common import (
   operator_contact, FinelyDividedTaskIterator, 
@@ -38,11 +44,23 @@ __all__ = [
   "create_spatial_index_mesh_tasks",
   "create_unsharded_multires_mesh_tasks",
   "create_sharded_multires_mesh_tasks",
+  "create_sharded_multires_mesh_from_unsharded_tasks",
+  "create_xfer_meshes_tasks",
 ]
 
 # split the work up into ~1000 tasks (magnitude 3)
 def create_mesh_manifest_tasks(layer_path, magnitude=3, mesh_dir=None):
   assert int(magnitude) == magnitude
+  assert magnitude >= 0
+
+  protocol = cloudfiles.paths.get_protocol(layer_path)
+  if protocol == "file":
+    return [
+      partial(MeshManifestFilesystemTask, 
+        layer_path=layer_path,
+        mesh_dir=mesh_dir,
+      )
+    ]
 
   start = 10 ** (magnitude - 1)
   end = 10 ** magnitude
@@ -52,11 +70,19 @@ def create_mesh_manifest_tasks(layer_path, magnitude=3, mesh_dir=None):
       return (10 ** magnitude) - 1
     def __iter__(self):
       for prefix in range(1, start):
-        yield MeshManifestTask(layer_path=layer_path, prefix=str(prefix) + ':', mesh_dir=mesh_dir)
+        yield partial(MeshManifestPrefixTask, 
+          layer_path=layer_path, 
+          prefix=str(prefix) + ':', 
+          mesh_dir=mesh_dir
+        )
 
       # enumerate from e.g. 100 to 999
       for prefix in range(start, end):
-        yield MeshManifestTask(layer_path=layer_path, prefix=prefix, mesh_dir=mesh_dir)
+        yield partial(MeshManifestPrefixTask, 
+          layer_path=layer_path, 
+          prefix=str(prefix),
+          mesh_dir=mesh_dir
+        )
 
   return MeshManifestTaskIterator()
 
@@ -67,7 +93,7 @@ def create_meshing_tasks(
     mesh_dir=None, cdn_cache=False, dust_threshold=None,
     object_ids=None, progress=False, fill_missing=False,
     encoding='precomputed', spatial_index=True, sharded=False,
-    compress='gzip'
+    compress='gzip', closed_dataset_edges=True
   ):
   shape = Vec(*shape)
 
@@ -112,6 +138,7 @@ def create_meshing_tasks(
         spatial_index=spatial_index,
         sharded=sharded,
         compress=compress,
+        closed_dataset_edges=closed_dataset_edges,
       )
 
     def on_finish(self):
@@ -132,6 +159,7 @@ def create_meshing_tasks(
           'spatial_index': spatial_index,
           'sharded': sharded,
           'compress': compress,
+          'closed_dataset_edges': closed_dataset_edges,
         },
         'by': operator_contact(),
         'date': strftime('%Y-%m-%d %H:%M %Z'),
@@ -230,7 +258,7 @@ def create_graphene_hybrid_mesh_manifest_tasks(
       return len(prefixes)
     def __iter__(self):
       for prefix in prefixes:
-        yield MeshManifestTask(layer_path=cloudpath, prefix=str(prefix))
+        yield partial(MeshManifestPrefixTask, layer_path=cloudpath, prefix=str(prefix))
 
   return GrapheneHybridMeshManifestTaskIterator()
 
@@ -387,6 +415,152 @@ def create_unsharded_multires_mesh_tasks(
 
   return UnshardedMultiResTaskIterator()
 
+def create_xfer_meshes_tasks(
+  src:str,
+  dest:str,
+  mesh_dir:Optional[str] = None, 
+  magnitude=2,
+):
+  cv_src = CloudVolume(src)
+  cf_dest = CloudFiles(dest)
+
+  if not mesh_dir:
+    info = cf_dest.get_json("info")
+    if info.get("mesh", None):
+      mesh_dir = info.get("mesh")
+
+  cf_dest.put_json(f"{mesh_dir}/info", cv_src.mesh.meta.info)
+
+  alphabet = [ str(i) for i in range(10) ]
+  if cv_src.mesh.meta.is_sharded():
+    alphabet += [ 'a', 'b', 'c', 'd', 'e', 'f' ]
+
+  prefixes = itertools.product(*([ alphabet ] * magnitude))
+  prefixes = [ "".join(x) for x in prefixes ]
+
+  # explicitly enumerate all prefixes smaller than the magnitude.
+  for i in range(1, magnitude):
+    explicit_prefix = itertools.product(*([ alphabet ] * i))
+    explicit_prefix = [ "".join(x) for x in explicit_prefix ]
+    if cv_src.mesh.meta.is_sharded():
+      prefixes += [ f"{x}." for x in explicit_prefix ]
+    else:
+      prefixes += [ f"{x}:0" for x in explicit_prefix ]
+
+  return [
+    partial(TransferMeshFilesTask,
+      src=src,
+      dest=dest,
+      prefix=prefix,
+      mesh_dir=mesh_dir,
+    )
+    for prefix in prefixes
+  ]
+
+def create_sharded_multires_mesh_from_unsharded_tasks(
+  src:str, 
+  dest:str,
+  shard_index_bytes=2**13, 
+  minishard_index_bytes=2**15,
+  num_lod:int = 1, 
+  draco_compression_level:int = 1,
+  vertex_quantization_bits:int = 16,
+  minishard_index_encoding="gzip", 
+  mesh_dir:Optional[str] = None, 
+) -> Iterator[MultiResShardedMeshMergeTask]: 
+  
+  configure_multires_info(
+    dest, 
+    vertex_quantization_bits, 
+    mesh_dir
+  )
+
+  cv_src = CloudVolume(src)
+  cf = CloudFiles(cv_src.mesh.meta.layerpath)
+
+  all_labels = []
+  SEGID_RE = re.compile(r'(\d+):0(?:\.gz|\.br|\.zstd)?$')
+  for path in cf.list():
+    match = SEGID_RE.search(path)
+    if match is None:
+      continue
+    (segid,) = match.groups()
+    all_labels.append(int(segid))
+
+  (shard_bits, minishard_bits, preshift_bits) = \
+    compute_shard_params_for_hashed(
+      num_labels=len(all_labels),
+      shard_index_bytes=int(shard_index_bytes),
+      minishard_index_bytes=int(minishard_index_bytes),
+    )
+
+  cv_dest = CloudVolume(dest, mesh_dir=mesh_dir)
+  cv_dest.mesh.meta.info["mip"] = cv_src.mesh.meta.mip
+  cv_dest.commit_info()
+
+  spec = ShardingSpecification(
+    type='neuroglancer_uint64_sharded_v1',
+    preshift_bits=preshift_bits,
+    hash='murmurhash3_x86_128',
+    minishard_bits=minishard_bits,
+    shard_bits=shard_bits,
+    minishard_index_encoding=minishard_index_encoding,
+    data_encoding="raw", # draco encoded meshes
+  )
+
+  cv_dest.mesh.meta.info['sharding'] = spec.to_dict()
+  cv_dest.mesh.meta.commit_info()
+
+  cv_dest = CloudVolume(dest, mesh_dir=mesh_dir)
+
+  # perf: ~66.5k hashes/sec on M1 ARM64
+  shardfn = lambda lbl: cv_dest.mesh.reader.spec.compute_shard_location(lbl).shard_number
+
+  shard_labels = defaultdict(list)
+  for label in tqdm(all_labels, desc="Hashes"):
+    shard_labels[shardfn(label)].append(label)
+  del all_labels
+
+  cf = CloudFiles(cv_dest.mesh.meta.layerpath, progress=True)
+  files = ( 
+    (str(shardno) + '.labels', labels) 
+    for shardno, labels in shard_labels.items() 
+  )
+  cf.put_jsons(
+    files, compress="gzip", 
+    cache_control="no-cache", total=len(shard_labels)
+  )
+
+  cv_dest.provenance.processing.append({
+    'method': {
+      'task': 'MultiResShardedFromUnshardedMeshMergeTask',
+      'src': src,
+      'dest': dest,
+      'num_lod': num_lod,
+      'vertex_quantization_bits': vertex_quantization_bits,
+      'preshift_bits': preshift_bits, 
+      'minishard_bits': minishard_bits, 
+      'shard_bits': shard_bits,
+      'mesh_dir': mesh_dir,
+      'draco_compression_level': draco_compression_level,
+    },
+    'by': operator_contact(),
+    'date': strftime('%Y-%m-%d %H:%M %Z'),
+  }) 
+  cv_dest.commit_provenance()
+
+  return [
+    partial(MultiResShardedFromUnshardedMeshMergeTask,
+      src=src, 
+      dest=dest, 
+      shard_no=shard_no, 
+      num_lod=num_lod,
+      mesh_dir=mesh_dir, 
+      draco_compression_level=draco_compression_level,
+    )
+    for shard_no in shard_labels.keys()
+  ]
+
 def create_sharded_multires_mesh_tasks(
   cloudpath:str, 
   shard_index_bytes=2**13, 
@@ -442,7 +616,7 @@ def create_sharded_multires_mesh_tasks(
     shard_labels[shardfn(label)].append(label)
   del all_labels
 
-  cf = CloudFiles(cv.skeleton.meta.layerpath, progress=True)
+  cf = CloudFiles(cv.mesh.meta.layerpath, progress=True)
   files = ( 
     (str(shardno) + '.labels', labels) 
     for shardno, labels in shard_labels.items() 
