@@ -56,7 +56,7 @@ class Tuple2(click.ParamType):
 
 @click.group()
 @click.option("-p", "--parallel", default=1, help="Run with this number of parallel processes. If 0, use number of cores.")
-@click.version_option(version="0.9.3")
+@click.version_option(version="2.0.0")
 @click.pass_context
 def main(ctx, parallel):
   """
@@ -359,6 +359,44 @@ def meshgroup():
   """
   pass
 
+@meshgroup.command("xfer")
+@click.argument("src")
+@click.argument("dest")
+@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
+@click.option("--sharded", is_flag=True, default=False, help="Generate shard fragments instead of outputing mesh fragments.", show_default=True)
+@click.option("--dir", "mesh_dir", type=str, default=None, help="Write meshes into this directory instead of the one indicated in the info file.")
+@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks.", show_default=True)
+@click.pass_context
+def mesh_xfer(
+  ctx, src, dest, queue,
+  sharded, mesh_dir, magnitude
+):
+  """
+  Transfer meshes to another location.
+  Enables conversion of unsharded to sharded
+  as well.
+  """
+  src = cloudfiles.paths.normalize(src)
+  dest = cloudfiles.paths.normalize(dest)
+  cv_src = CloudVolume(src)
+
+  if not cv_src.mesh.meta.is_sharded() and sharded:
+    tasks = tc.create_sharded_multires_mesh_from_unsharded_tasks(
+      src, dest,
+      num_lod=1, 
+      mesh_dir=mesh_dir, 
+    )
+  else:
+    tasks = tc.create_xfer_meshes_tasks(
+      src, dest, 
+      mesh_dir=mesh_dir, 
+      magnitude=magnitude,
+    )
+
+  parallel = int(ctx.obj.get("parallel", 1))
+  tq = TaskQueue(normalize_path(queue))
+  tq.insert(tasks, parallel=parallel)
+
 @meshgroup.command("forge")
 @click.argument("path")
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@@ -372,12 +410,13 @@ def meshgroup():
 @click.option('--compress', default="gzip", help="Set the image compression scheme. Options: 'none', 'gzip', 'br'", show_default=True)
 @click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.", show_default=True)
 @click.option('--sharded', is_flag=True, default=False, help="Generate shard fragments instead of outputing mesh fragments.", show_default=True)
+@click.option('--closed-edge/--open-edge', is_flag=True, default=True, help="Whether meshes are closed on the side that contacts the dataset boundary.", show_default=True)
 @click.pass_context
 def mesh_forge(
   ctx, path, queue, mip, shape, 
   simplify, fill_missing, max_error, 
   dust_threshold, dir, compress, 
-  spatial_index, sharded
+  spatial_index, sharded, closed_edge
 ):
   """
   (1) Synthesize meshes from segmentation cutouts.
@@ -404,7 +443,7 @@ def mesh_forge(
     mesh_dir=dir, cdn_cache=False, dust_threshold=dust_threshold,
     object_ids=None, progress=False, fill_missing=fill_missing,
     encoding='precomputed', spatial_index=spatial_index, 
-    sharded=sharded, compress=compress
+    sharded=sharded, compress=compress, closed_dataset_edges=closed_edge
   )
 
   parallel = int(ctx.obj.get("parallel", 1))
